/*
 *
 * Copyright (c) 2003 by FORCE Computers
 *
 * Note that this file is based on parts of OpenIPMI
 * written by Corey Minyard <minyard@mvista.com>
 * of MontaVista Software. Corey's code was helpful
 * and many thanks go to him. He gave the permission
 * to use this code in OpenHPI under BSD license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *     Thomas Kanngieser <thomas.kanngieser@fci.com>
 */

#ifndef dIpmiDomain_h
#define dIpmiDomain_h


#include <stdlib.h>
#include <string.h>


__BEGIN_DECLS
#include "SaHpi.h"
#include <openhpi.h>
#include <epath_utils.h>
#include <uid_utils.h>
__END_DECLS


#ifndef dIpmiCon_h
#include "ipmi_con.h"
#endif

#ifndef dIpmiMc_h
#include "ipmi_mc.h"
#endif

#ifndef dIpmiEntity_h
#include "ipmi_entity.h"
#endif

#ifndef dIpmiSensor_h
#include "ipmi_sensor.h"
#endif


#define dMaxMcMissedResponses 1

/* Re-query the SEL every 10 seconds by default. */
#define dIpmiSelQueryInterval 10000

class cIpmiEventAsync
{
public:
  cIpmiEvent m_event;

  cIpmiEventAsync( const cIpmiEvent &event )
    : m_event( event )  {}
};


class cIpmiDomain
{
protected:
  cIpmiCon     *m_con;

  int           m_is_atca;

  // The main set of SDRs on a BMC.
  cIpmiSdrs    *m_main_sdrs;

  // The sensors that came from the main SDR.
  cIpmiSensor **m_sensors_in_main_sdr;
  unsigned int  m_sensors_in_main_sdr_count;

  // Major and minor versions of the connection.
  unsigned int  m_major_version;
  unsigned int  m_minor_version;
  bool          m_sdr_repository_support;

  // A special MC used to represent the system interface.
  cIpmiMc      *m_si_mc;

  GList        *m_mcs; // list of all msc

  cIpmiEntityInfo *m_entities;

  unsigned int  m_default_sel_rescan_time;

  GList        *m_sel_scan_timer;

  int CheckAtca();
  int GetChannels();

  enum tScanState
  {
    eScanOk,
    eScanStateRescan,
    eScanStateError
  };

  cIpmiMc *Scan( const cIpmiAddr &addr, int &missed_responses, tScanState &state );
  void SelTimerAddToList( cIpmiMc *mc );

  cIpmiMc *NewMc( const cIpmiAddr &addr );
  bool     CleanupMc( cIpmiMc *mc );

public:
  cIpmiDomain();
  virtual ~cIpmiDomain();

  bool Init( cIpmiCon *c );
  void Cleanup();

  cIpmiEntityInfo &Entities() { return *m_entities; }
  
  cIpmiMc *FindMcByAddr( const cIpmiAddr &addr );
  cIpmiMc *FindOrCreateMcBySlaveAddr( unsigned int slave_addr );
  cIpmiMc *ScanMc( int channel, unsigned int addr );
  cIpmiMc *ScanSi( int channel );
  int      SendCommand( const cIpmiAddr &addr, const cIpmiMsg &msg, cIpmiMsg &rsp_msg );
  cIpmiSensor **GetSdrSensors( cIpmiMc *mc, unsigned int &count );
  void          SetSdrSensors( cIpmiMc *mc, cIpmiSensor **sensors,
                               unsigned int count );
  int GetEventRcvr();

  void SelTimerAdd( cIpmiMc *mc );
  void SelTimerRem( cIpmiMc *mc );
  void SelTimerCheck();

  void HandleHotswapEvent( cIpmiMc *mc, cIpmiSensor *sensor, cIpmiEvent *event );
  void HandleEvent( cIpmiMc *mc, cIpmiSensor *sensor, cIpmiEvent *event );

  void SystemHandleEvent( cIpmiMc *m, cIpmiEvent *event );

protected:
  GList *m_async_event_list;

public:
  void AddAsyncEvent( const cIpmiAddr &addr, const cIpmiMsg &msg );
  void HandleAsyncEvents();

  // called with a list of events to handle
  void HandleEvents( cIpmiMc *mc, GList *list );

  cIpmiEntity *VerifyEntity( cIpmiEntity *ent );
  cIpmiMc     *VerifyMc( cIpmiMc *mc );
  cIpmiSensor *VerifySensor( cIpmiSensor *s );

  void McRescan();

  virtual void AddHpiEvent( oh_event *event ) = 0;
  virtual GSList *GetHpiEventList() = 0;

  // called from IPMI framework to create HPI hotswap events
  virtual void IfHotswapEvent( cIpmiMc *mc, cIpmiSensor *sensor,
                               cIpmiEvent *event ) = 0;

  // called from IPMI framework to create/destroy HPI resources
  virtual void IfEntityAdd( cIpmiEntity *ent ) = 0;
  virtual void IfEntityRem( cIpmiEntity *ent ) = 0;

  // called from IPMI framework to create/destroy HPI rdr
  virtual void IfSensorAdd( cIpmiEntity *ent, cIpmiSensor *sensor ) = 0;
  virtual void IfSensorRem( cIpmiEntity *ent, cIpmiSensor *sensor ) = 0;
  virtual void IfSensorDiscreteEvent( cIpmiSensor  *sensor,
                                      tIpmiEventDir dir,
                                      int           offset,
                                      int           severity,
                                      int           prev_severity,
                                      cIpmiEvent   *event ) = 0;
  virtual void IfSensorThresholdEvent( cIpmiSensor        *sensor,
                                       tIpmiEventDir      dir,
                                       tIpmiThresh        threshold,
                                       tIpmiEventValueDir high_low,
                                       cIpmiEvent        *event ) = 0;

  virtual const char *EntityRoot() = 0;
  virtual oh_handler_state *GetHandler() = 0;

  virtual SaHpiRptEntryT *FindResource( SaHpiResourceIdT id ) = 0;
};


#endif
