/*
 * ipmi_sel.h
 *
 * Copyright (c) 2003 by FORCE Computers
 *
 * Note that this file is based on parts of OpenIPMI
 * written by Corey Minyard <minyard@mvista.com>
 * of MontaVista Software. Corey's code was helpful
 * and many thanks go to him. He gave the permission
 * to use this code in OpenHPI under BSD license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *     Thomas Kanngieser <thomas.kanngieser@fci.com>
 */


#ifndef dIpmiSel_h
#define dIpmiSel_h


#ifndef dIpmiEvent_h
#include "ipmi_event.h"
#endif

#include "glib.h"


#define dMaxSelFetchRetries 3


class cIpmiSel
{
protected:
  cIpmiMc      *m_mc;

  // LUN we are attached with.
  int           m_lun;

  unsigned char m_major_version;
  unsigned char m_minor_version;
  unsigned char m_entries;
  unsigned int  m_last_addition_timestamp;
  unsigned int  m_last_erase_timestamp;
  bool          m_overflow;
  bool          m_supports_delete_sel;
  bool          m_supports_partial_add_sel;
  bool          m_supports_reserve_sel;
  bool          m_supports_get_sel_allocation;

  bool          m_fetched;

  // When fetching the data in event-driven mode, these are the
  // variables that track what is going on.
  unsigned int  m_reservation;
  bool          m_sels_changed;

  // This is the actual list of events and the number of non-deleted
  // events and the number of deleted events.  Note that events may
  // contain more items than num_sels, num_sels only counts the
  // number of non-deleted events in the list.  del_sels+num_sels
  // should be the number of events.

  // Holds an event in the list of events.
  class cIpmiSelEventHolder
  {
  public:
    bool       m_deleted;
    cIpmiEvent m_event;
  };

  GList        *m_events;
  unsigned int  m_num_sels;
  unsigned int  m_del_sels;

  enum tReadRecord
  {
    eReadOk,
    eReadEndOfSel,
    eReadReservationLost,
    eReadError
  };

  int GetInfo( unsigned int &fetched_num_sels );
  int Reserve();

  void FreeEvents();
  void FreeDeletedEvents();
  int SelClear( cIpmiMc *mc );
  cIpmiSelEventHolder *FindEvent( unsigned int recid );
  GList *ClearSel( GList *new_events );

public:
  cIpmiSel( cIpmiMc *mc, unsigned int lun );
  ~cIpmiSel();

  GList *GetEvents();
  int AddEvent( cIpmiEvent *new_event );
};


#endif
