/*      -*- linux-c -*-
 *
 *      $Id: snmp_bc.c,v 1.59 2003/12/10 21:23:22 djudkovi Exp $
 *
 * (C) Copyright IBM Corp. 2003
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *      Renier Morales <renierm@users.sf.net>
 *
 */

#include <snmp_bc.h>
#include <rpt_utils.h>
#include <epath_utils.h>

/**
 * snmp_bc_get_event:
 * @hnd: 
 * @event: 
 * @timeout: 
 *
 * Return value: 
 **/
static int snmp_bc_get_event(void *hnd, struct oh_event *event, struct timeval *timeout)
{
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        
        if(g_slist_length(handle->eventq)>0) {
                memcpy(event, handle->eventq->data, sizeof(*event));
                free(handle->eventq->data);
                handle->eventq = g_slist_remove_link(handle->eventq, handle->eventq);
                return 1;
        } else {
                return 0;
	}
}

#define find_sensors(rdr_array) \
do { \
        int j; \
	for(j=0; rdr_array[j].mib.oid != NULL; j++) { \
		e = snmp_bc_discover_sensors(custom_handle->ss, \
                                             parent_ep, \
                                             &rdr_array[j]); \
                if(e != NULL) { \
                        struct SensorMibInfo *rdr_mib = g_memdup(&(rdr_array[j].mib), \
                                                                 sizeof(struct SensorMibInfo)); \
                        oh_add_rdr(handle->rptcache,rid,&(e->u.rdr_event.rdr), rdr_mib); \
                        handle->eventq = g_slist_append(handle->eventq, e); \
                } \
        } \
} while(0)

#define find_controls(rdr_array) \
do { \
        int j; \
        for(j=0; rdr_array[j].mib.oid != NULL; j++) { \
	        e = snmp_bc_discover_controls(custom_handle->ss, \
                                              parent_ep, \
                                              &rdr_array[j]); \
                if(e != NULL) { \
                        struct ControlMibInfo *rdr_mib = g_memdup(&(rdr_array[j].mib), \
                                                                  sizeof(struct ControlMibInfo)); \
                        oh_add_rdr(handle->rptcache,rid,&(e->u.rdr_event.rdr), rdr_mib); \
                        handle->eventq = g_slist_append(handle->eventq, e); \
                } \
        } \
} while(0)



#define find_inventories(rdr_array) \
do { \
        int j; \
        for (j=0; rdr_array[j].mib.oid.OidManufacturer != NULL; j++) { \
                e = snmp_bc_discover_inventories(custom_handle->ss, \
                                              parent_ep, \
                                              &rdr_array[j]); \
                if(e != NULL) { \
                        struct InventoryMibInfo *rdr_mib = g_memdup(&(rdr_array[j].mib), \
                                                                    sizeof(struct InventoryMibInfo)); \
                        oh_add_rdr(handle->rptcache,rid,&(e->u.rdr_event.rdr), rdr_mib); \
                        handle->eventq = g_slist_append(handle->eventq, e); \
                } \
        } \
} while(0)


static int snmp_bc_discover_resources(void *hnd)
{
        SaHpiEntityPathT entity_root;        
        guint i;
        struct oh_event *e;
	struct snmp_value get_value, get_active;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        char *root_tuple = (char *)g_hash_table_lookup(handle->config,"entity_root");        
                
        memset(&entity_root, 0 , sizeof(SaHpiEntityPathT));
        string2entitypath(root_tuple, &entity_root);
        append_root(&entity_root);

	/* Blade vector gives us info for chassis, subchassis, blade, 
	   and blade daughter (add in) cards */
        if((snmp_get(custom_handle->ss,SNMP_BC_BLADE_VECTOR,&get_value)) == 0 &&
	   (get_value.type == ASN_OCTET_STR)) {
                e = snmp_bc_discover_chassis(get_value.string,&entity_root);
                if(e != NULL) {
                        struct ResourceMibInfo *res_mib =
                                g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_CHASSIS].mib),
                                         sizeof(struct snmp_rpt));
                        oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
                        handle->eventq = g_slist_append(handle->eventq, e);
                        SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                        SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;
			find_sensors(snmp_bc_chassis_sensors);                        
			find_controls(snmp_bc_chassis_controls);
			find_inventories(snmp_bc_chassis_inventories);
                }

                for(i=0; i < BC_MAX_SUBCHASSIS; i++) {
                        e = snmp_bc_discover_subchassis(get_value.string, &entity_root,	i);
                        if(e != NULL) {
                                struct ResourceMibInfo *res_mib =
                                        g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_SUBCHASSIS].mib),
                                                 sizeof(struct snmp_rpt));
                                oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
                                handle->eventq = g_slist_append(handle->eventq, e);
                                SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                                SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;
				find_sensors(snmp_bc_subchassis_sensors);
				find_controls(snmp_bc_subchassis_controls);
				find_inventories(snmp_bc_subchassis_inventories);
                        }
                }

                for(i=0; i < strlen(get_value.string); i++) {
                        e = snmp_bc_discover_blade(get_value.string, &entity_root, i);
                        if(e != NULL) {
                                struct ResourceMibInfo *res_mib =
                                        g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_BLADE].mib),
                                                 sizeof(struct snmp_rpt));
                                oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
                                handle->eventq = g_slist_append(handle->eventq, e);
                                SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                                SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;
				find_sensors(snmp_bc_blade_sensors);
				find_controls(snmp_bc_blade_controls);
				find_inventories(snmp_bc_blade_inventories);
                        }
			
			e = snmp_bc_discover_blade_addin(custom_handle->ss, get_value.string,
                                                         &entity_root, i);
			if(e != NULL) {
                                struct ResourceMibInfo *res_mib =
                                        g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_BLADE_ADDIN_CARD].mib),
                                                 sizeof(struct snmp_rpt));
                                oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
				handle->eventq = g_slist_append(handle->eventq, e);
                                SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                                SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;
				find_sensors(snmp_bc_blade_addin_sensors);
				find_controls(snmp_bc_blade_addin_controls);
				find_inventories(snmp_bc_blade_addin_inventories);
			}
                }
        } else {
                /* If we get here, something is hosed. No need to do more discovery */
                dbg("Couldn't fetch SNMP %s vector; Type=%d\n",SNMP_BC_BLADE_VECTOR,get_value.type);
                dbg("There is either no chassis or the management modules are down.");
                return -1;
        }
        
        if((snmp_get(custom_handle->ss,SNMP_BC_FAN_VECTOR,&get_value)) == 0 &&
	   (get_value.type == ASN_OCTET_STR)) {
                for(i=0; i < strlen(get_value.string); i++) {
                        e = snmp_bc_discover_fan(get_value.string, &entity_root, i);
                        if(e != NULL) {
                                struct ResourceMibInfo *res_mib =
                                        g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_BLOWER_MODULE].mib),
                                                 sizeof(struct snmp_rpt));
                                oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
                                handle->eventq = g_slist_append(handle->eventq, e);
                                SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                                SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;
				find_sensors(snmp_bc_fan_sensors);
				find_controls(snmp_bc_fan_controls);
				find_inventories(snmp_bc_fan_inventories);
                        }
                }
        } else {
		dbg("Couldn't fetch SNMP %s vector; Type=%d\n",SNMP_BC_FAN_VECTOR,get_value.type);
                dbg("There is either no chassis or the management modules are down.");
                return -1;
        }

        if((snmp_get(custom_handle->ss,SNMP_BC_POWER_VECTOR,&get_value)) == 0 &&
	   (get_value.type == ASN_OCTET_STR)) {
                for(i=0; i < strlen(get_value.string); i++) {
                        e = snmp_bc_discover_power(get_value.string, &entity_root, i);
                        if(e != NULL) {
                                struct ResourceMibInfo *res_mib =
                                        g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_POWER_MODULE].mib),
                                                 sizeof(struct snmp_rpt));
                                oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
                                handle->eventq = g_slist_append(handle->eventq, e);
                                SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                                SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;
                                find_sensors(snmp_bc_power_sensors);
				find_controls(snmp_bc_power_controls);
				find_inventories(snmp_bc_power_inventories);
                        }
                }
        } else {
		dbg("Couldn't fetch SNMP %s vector; Type=%d\n",SNMP_BC_POWER_VECTOR,get_value.type);
                dbg("There is either no chassis or the management modules are down.");
                return -1;
        }

        if((snmp_get(custom_handle->ss,SNMP_BC_SWITCH_VECTOR,&get_value)) == 0 &&
	   (get_value.type == ASN_OCTET_STR)) {
                for(i=0; i < strlen(get_value.string); i++) {
                        e = snmp_bc_discover_switch(get_value.string, &entity_root, i);
                        if(e != NULL) {
                                struct ResourceMibInfo *res_mib =
                                        g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_SWITCH_MODULE].mib),
                                                 sizeof(struct snmp_rpt));
                                oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
                                handle->eventq = g_slist_append(handle->eventq, e);
                                SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                                SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;
				find_sensors(snmp_bc_switch_sensors);
				find_controls(snmp_bc_switch_controls);
				find_inventories(snmp_bc_switch_inventories);
                         }
                }
        } else {
		dbg("Couldn't fetch SNMP %s vector; Type=%d\n",SNMP_BC_SWITCH_VECTOR,get_value.type);
                dbg("There is either no chassis or the management modules are down.");
                return -1;
        }

        if((snmp_get(custom_handle->ss,SNMP_BC_MEDIATRAY_EXISTS,&get_value)) == 0 &&
	   (get_value.type == ASN_INTEGER)) {
		e = snmp_bc_discover_mediatray(get_value.integer, &entity_root, 0);
		if(e != NULL) {
                        struct ResourceMibInfo *res_mib =
                                        g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_MEDIA_TRAY].mib),
                                                 sizeof(struct snmp_rpt));
                        oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
			handle->eventq = g_slist_append(handle->eventq, e);
                        SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                        SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;                        
			find_sensors(snmp_bc_mediatray_sensors);
			find_controls(snmp_bc_mediatray_controls);
			find_inventories(snmp_bc_mediatray_inventories);
		}
        } else {
		dbg("Couldn't fetch SNMP %s vector; Type=%d\n",SNMP_BC_MEDIATRAY_EXISTS,get_value.type);
                dbg("There is either no chassis or the management modules are down.");
                return -1;
        }

        if((snmp_get(custom_handle->ss,SNMP_BC_MGMNT_VECTOR,&get_value)) == 0 && 
	   (get_value.type == ASN_OCTET_STR)) {
                for(i=0; i < strlen(get_value.string); i++) {
			if((snmp_get(custom_handle->ss,SNMP_BC_MGMNT_ACTIVE,&get_active)) == 0 &&
			    (get_active.type == ASN_INTEGER) && (get_active.integer == i+1)) { 	
				e = snmp_bc_discover_mgmnt(get_value.string, &entity_root, i);
				if(e != NULL) {
                                        struct ResourceMibInfo *res_mib =
                                            g_memdup(&(snmp_rpt_array[BC_RPT_ENTRY_MGMNT_MODULE].mib),
                                                     sizeof(struct snmp_rpt));
                                        oh_add_resource(handle->rptcache,&(e->u.res_event.entry),res_mib);
					handle->eventq = g_slist_append(handle->eventq, e);
                                        SaHpiResourceIdT rid = e->u.res_event.entry.ResourceId;
                                        SaHpiEntityPathT parent_ep = e->u.res_event.entry.ResourceEntity;                                        
					find_sensors(snmp_bc_mgmnt_sensors);
					find_controls(snmp_bc_mgmnt_controls);
					find_inventories(snmp_bc_mgmnt_inventories);
				}
			}
		}
	} else {
		dbg("Couldn't fetch SNMP %s vector; Type=%d\n",SNMP_BC_MGMNT_VECTOR,get_value.type);
                dbg("There is either no chassis or the management modules are down.");
                return -1;
        }

        return SA_OK;
}

static int snmp_bc_set_resource_tag(void *hnd, SaHpiResourceIdT id, SaHpiTextBufferT *tag)
{
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        SaHpiRptEntryT *resource = oh_get_resource_by_id(handle->rptcache, id);
        struct oh_event *e = NULL;
        guint datalength = tag->DataLength;

        if (!resource) {
                dbg("Error. Cannot set resource tag in plugin. No resource found by that id.");
                return SA_ERR_HPI_NOT_PRESENT;
        }

        if (datalength > SAHPI_MAX_TEXT_BUFFER_LENGTH) datalength = SAHPI_MAX_TEXT_BUFFER_LENGTH;

        strncpy(resource->ResourceTag.Data, tag->Data, datalength);
        resource->ResourceTag.DataLength = tag->DataLength;
        resource->ResourceTag.DataType = tag->DataType;
        resource->ResourceTag.Language = tag->Language;

        /** Can we persist this to hardware or disk? */

        /* Add changed resource to event queue */
        e = g_malloc0(sizeof(struct oh_event));
        e->type = OH_ET_RESOURCE;
        e->u.res_event.entry = *resource;

        handle->eventq = g_slist_append(handle->eventq, e);
        
        return SA_OK;
}

static int snmp_bc_set_resource_severity(void *hnd, SaHpiResourceIdT id, SaHpiSeverityT sev)
{
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        SaHpiRptEntryT *resource = oh_get_resource_by_id(handle->rptcache, id);
        struct oh_event *e = NULL;

        if (!resource) {
                dbg("Error. Cannot set resource severity in plugin. No resource found by that id.");
                return SA_ERR_HPI_NOT_PRESENT;
        }

        resource->ResourceSeverity = sev;

        /** Can we persist this to disk? */

        /* Add changed resource to event queue */
        e = g_malloc0(sizeof(struct oh_event));
        e->type = OH_ET_RESOURCE;
        e->u.res_event.entry = *resource;
        
        handle->eventq = g_slist_append(handle->eventq, e);

        return SA_OK;
}

/**
 * snmp_bc_get_self_id:
 * @hnd: 
 * @id: 
 * 
 * Return value: 
 **/
static int snmp_bc_get_self_id(void *hnd, SaHpiResourceIdT id)
{
        return -1;
}

static int snmp_bc_get_sensor_data(void *hnd, SaHpiResourceIdT id,
                                   SaHpiSensorNumT num,
                                   SaHpiSensorReadingT *data)
{
        gchar *oid;
	SaHpiSensorReadingT working;
        struct snmp_value get_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRdrT *rdr = oh_get_rdr_by_type(handle->rptcache, id, SAHPI_SENSOR_RDR, num);
        struct SensorMibInfo *s =
                (struct SensorMibInfo *)oh_get_rdr_data(handle->rptcache, id, rdr->RecordId);
        
	memset(&working, 0, sizeof(SaHpiSensorReadingT));

        /* Extract index from rdr id and get the snmp of the sensor */
        oid = snmp_derive_objid(rdr->Entity, s->oid);
	if(oid == NULL) {
		dbg("NULL SNMP OID returned for %s\n",s->oid);
		return -1;
	}

        /* Read the sensor value */
        if(snmp_get(custom_handle->ss, oid, &get_value) != 0){
                dbg("SNMP could not read sensor %s. Type = %d",oid,get_value.type);
		g_free(oid);
                return SA_ERR_HPI_NO_RESPONSE;
        }
        g_free(oid);

        /* Based on the sensor description, construct a reading to send up */
        /* format the value into the reading for each type of reading format */
        working.ValuesPresent = rdr->RdrTypeUnion.SensorRec.DataFormat.ReadingFormats;
        if(working.ValuesPresent & SAHPI_SRF_RAW) {
                if(get_value.type != ASN_INTEGER) {
                        dbg("Sensor value type mismatches reading format.");
                        return -1;
                } else {
                        working.Raw = (SaHpiUint32T)get_value.integer;
                }
        }

        if(working.ValuesPresent & SAHPI_SRF_INTERPRETED) {
                if(get_value.type == ASN_INTEGER) {
                        working.Interpreted.Type = SAHPI_SENSOR_INTERPRETED_TYPE_INT32;
                        working.Interpreted.Value.SensorInt32 = get_value.integer;
                } else {
			SaHpiSensorInterpretedUnionT value;
			
		       	working.Interpreted.Type = rdr->RdrTypeUnion.SensorRec.DataFormat.Range.Max.Interpreted.Type;
			if(rdr->RdrTypeUnion.SensorRec.DataFormat.Range.Max.Interpreted.Type == SAHPI_SENSOR_INTERPRETED_TYPE_BUFFER) {
				strncpy(working.Interpreted.Value.SensorBuffer,
					get_value.string,
                                                SAHPI_SENSOR_BUFFER_LENGTH);
			} else {
				if(s->convert_snmpstr >= 0) {
					if(get_interpreted_value(get_value.string,s->convert_snmpstr,&value)) {
						dbg("Error: get_interpreted_value for %s, (%s)\n",s->oid,get_value.string);
						return -1;
					}
					working.Interpreted.Value = value;
				} else {
					dbg("Sensor %s SNMP string value needs to be converted\n", s->oid);
					return -1;
				}
			}
                }
        }

        /** FIXME: Need to map events */
        if(working.ValuesPresent & SAHPI_SRF_EVENT_STATE) { 
                if(get_value.type == ASN_OCTET_STR) {
                        dbg("Do not know how to format strings as events yet.");
                } else {
                        SaHpiUint32T shifting = 1;
                        shifting = shifting << get_value.integer;
                        if(rdr->RdrTypeUnion.SensorRec.Events & shifting) {
                                working.EventStatus.SensorStatus = SAHPI_SENSTAT_SCAN_ENABLED |
                                                                 SAHPI_SENSTAT_EVENTS_ENABLED;
                                working.EventStatus.EventStatus = shifting;
                        }
                }
        }

	memcpy(data,&working,sizeof(SaHpiSensorReadingT));
        
        return SA_OK;
}

#define get_raw_thresholds(thdmask, thdoid, thdname) \
do { \
        if(rdr->RdrTypeUnion.SensorRec.ThresholdDefn.ReadThold & thdmask) { \
	        if(s->threshold_oids.RawThresholds.thdoid != NULL && s->threshold_oids.RawThresholds.thdoid[0] != '\0') { \
	                oid = snmp_derive_objid(rdr->Entity,s->threshold_oids.RawThresholds.thdoid); \
                        if(oid == NULL) { \
                                 dbg("NULL SNMP OID returned for %s\n",s->threshold_oids.RawThresholds.thdoid); \
                                 return -1; \
                        } \
	                if((snmp_get(custom_handle->ss, oid, &get_value) != 0) | \
	                   (get_value.type != ASN_INTEGER)) { \
		                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
		                g_free(oid); \
		                return SA_ERR_HPI_NO_RESPONSE; \
	                } \
	                g_free(oid); \
	                found_raw++; \
	                working.thdname.Raw = get_value.integer; \
	                working.thdname.ValuesPresent = working.thdname.ValuesPresent | SAHPI_SRF_RAW; \
	        } else { \
		        dbg("Raw threshold defined as readable but no OID defined\n"); \
	        } \
        } \
} while(0)

#define get_interpreted_thresholds(thdmask, thdoid, thdname) \
do { \
        if(rdr->RdrTypeUnion.SensorRec.ThresholdDefn.ReadThold & thdmask) { \
	        if(s->threshold_oids.InterpretedThresholds.thdoid != NULL && s->threshold_oids.InterpretedThresholds.thdoid[0] != '\0') { \
		        oid = snmp_derive_objid(rdr->Entity,s->threshold_oids.InterpretedThresholds.thdoid); \
                        if(oid == NULL) { \
                                dbg("NULL SNMP OID returned for %s\n",s->threshold_oids.InterpretedThresholds.thdoid); \
                                return -1; \
                        } \
	         	if((snmp_get(custom_handle->ss, oid, &get_value) != 0) | \
	                   !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
			        dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
			        g_free(oid); \
			        return SA_ERR_HPI_NO_RESPONSE; \
		        } \
		        found_interpreted++; \
		        /* Means we always need to define this field in bc_resources.h */ \
		        working.thdname.Interpreted.Type = rdr->RdrTypeUnion.SensorRec.DataFormat.Range.Max.Interpreted.Type; \
		        working.thdname.ValuesPresent = working.thdname.ValuesPresent | SAHPI_SRF_INTERPRETED; \
		        if(get_value.type == ASN_INTEGER) { \
			         working.thdname.Interpreted.Value.SensorInt32 = get_value.integer; \
		        } else if(get_value.type == ASN_OCTET_STR && s->convert_snmpstr >= 0) { \
			        if(get_interpreted_value(get_value.string,s->convert_snmpstr,&value)) { \
				        dbg("Error: bad return from get_interpreted_value for %s\n",oid); \
                                        g_free(oid); \
				        return -1; \
			        } \
			        working.thdname.Interpreted.Value = value; \
		        } else { \
			        dbg("%s threshold is string but no conversion defined\n",oid); \
                                g_free(oid); \
			        return -1; \
		        } \
                        g_free(oid); \
	        } else { \
		        dbg("Interpreted threshold defined as readable but no OID defined\n"); \
	        } \
        } \
} while(0)

static int snmp_bc_get_sensor_thresholds(void *hnd, SaHpiResourceIdT id,
                                         SaHpiSensorNumT num,
                                         SaHpiSensorThresholdsT *thres)
{
        gchar *oid = NULL;
	int  found_raw, found_interpreted;
	SaHpiSensorThresholdsT working;
	SaHpiSensorInterpretedUnionT value;
        struct snmp_value get_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRdrT *rdr = oh_get_rdr_by_type(handle->rptcache, id, SAHPI_SENSOR_RDR, num);
        struct SensorMibInfo *s =
                (struct SensorMibInfo *)oh_get_rdr_data(handle->rptcache, id, rdr->RecordId);

        memset(&working, 0, sizeof(SaHpiSensorThresholdsT));

	if(rdr->RdrTypeUnion.SensorRec.ThresholdDefn.IsThreshold == SAHPI_TRUE) {
		found_raw = found_interpreted = 0;
		if(rdr->RdrTypeUnion.SensorRec.ThresholdDefn.TholdCapabilities & SAHPI_STC_RAW) {

			get_raw_thresholds(SAHPI_STM_LOW_MINOR, OidLowMinor, LowMinor);
			get_raw_thresholds(SAHPI_STM_LOW_MAJOR, OidLowMajor, LowMajor);
			get_raw_thresholds(SAHPI_STM_LOW_CRIT, OidLowCrit, LowCritical);
			get_raw_thresholds(SAHPI_STM_UP_MINOR, OidUpMinor, UpMinor);
			get_raw_thresholds(SAHPI_STM_UP_MAJOR, OidUpMajor, UpMajor);
			get_raw_thresholds(SAHPI_STM_UP_CRIT, OidUpCrit, UpCritical);

			/* FIXME:: Add PosThdHysteresis and NegThdHysteresis */			 	
		}

		if(rdr->RdrTypeUnion.SensorRec.ThresholdDefn.TholdCapabilities & SAHPI_STC_INTERPRETED) {

			get_interpreted_thresholds(SAHPI_STM_LOW_MINOR, OidLowMinor, LowMinor);
			get_interpreted_thresholds(SAHPI_STM_LOW_MAJOR, OidLowMajor, LowMajor);
			get_interpreted_thresholds(SAHPI_STM_LOW_CRIT, OidLowCrit, LowCritical);
			get_interpreted_thresholds(SAHPI_STM_UP_MINOR, OidUpMinor, UpMinor);
			get_interpreted_thresholds(SAHPI_STM_UP_MAJOR, OidUpMajor, UpMajor);
			get_interpreted_thresholds(SAHPI_STM_UP_CRIT, OidUpCrit, UpCritical);

			/* FIXME:: Add PosThdHysteresis and NegThdHysteresis */			 	
		}

		/* FIXME:: Do we need to add events as well? */

		if (found_raw || found_interpreted) {
			memcpy(thres,&working,sizeof(SaHpiSensorThresholdsT));
			return SA_OK;
		} else {
			dbg("No threshold values found\n");
			return -1;
		}
        } else {
                dbg("Thresholds requested, but sensor does not support them.\n");
                return SA_ERR_HPI_INVALID_CMD;
        }        
}

static int snmp_bc_set_sensor_thresholds(void *hnd, SaHpiResourceIdT id,
                                         SaHpiSensorNumT num,
                                         const SaHpiSensorThresholdsT *thres)
{
	/* Writable thresholds not supported */
        return SA_ERR_HPI_INVALID_CMD;
}

static int snmp_bc_get_sensor_event_enables(void *hnd, SaHpiResourceIdT id,
                                            SaHpiSensorNumT num,
                                            SaHpiSensorEvtEnablesT *enables)
{
        return -1;
}

static int snmp_bc_set_sensor_event_enables(void *hnd, SaHpiResourceIdT id,
                                            SaHpiSensorNumT num,
                                            const SaHpiSensorEvtEnablesT *enables)
{
        return -1;
}

static int snmp_bc_get_control_state(void *hnd, SaHpiResourceIdT id,
                                     SaHpiCtrlNumT num,
                                     SaHpiCtrlStateT *state)
{
        gchar *oid;
	int i, found;
	SaHpiCtrlStateT working;
        struct snmp_value get_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRdrT *rdr = oh_get_rdr_by_type(handle->rptcache, id, SAHPI_CTRL_RDR, num);
        struct ControlMibInfo *s =
                (struct ControlMibInfo *)oh_get_rdr_data(handle->rptcache, id, rdr->RecordId);

	memset(&working, 0, sizeof(SaHpiCtrlStateT));
	working.Type = rdr->RdrTypeUnion.CtrlRec.Type;
	
	switch (working.Type) {
	case SAHPI_CTRL_TYPE_DIGITAL:
		oid = snmp_derive_objid(rdr->Entity, s->oid);
		if(oid == NULL) {
			dbg("NULL SNMP OID returned for %s\n",s->oid);
			return -1;
		}
		if((snmp_get(custom_handle->ss, oid, &get_value) != 0) | (get_value.type != ASN_INTEGER)) {
			dbg("SNMP could not read %s; Type=%d.\n", oid, get_value.type);
			g_free(oid);
			return SA_ERR_HPI_NO_RESPONSE;
		}
		g_free(oid);
		
		found = 0;
		/* Icky dependency on SaHpiStateDigitalT enum */
		for(i=0; i<ELEMENTS_IN_SaHpiStateDigitalT; i++) {
			if(s->digitalmap[i] == get_value.integer) { 
				found++;
				break; 
			}
		}

		if(found) {
			switch (i) {
			case 0:
				working.StateUnion.Digital = SAHPI_CTRL_STATE_OFF;
				break;
			case 1:
				working.StateUnion.Digital = SAHPI_CTRL_STATE_ON;
				break;
			case 2:
				working.StateUnion.Digital = SAHPI_CTRL_STATE_PULSE_OFF;
				break;
			case 3:
				working.StateUnion.Digital = SAHPI_CTRL_STATE_PULSE_ON;
				break;
			case 4:
				working.StateUnion.Digital = SAHPI_CTRL_STATE_AUTO;
				break;
			default:
				dbg("Spec Change: MAX_SaHpiStateDigitalT incorrect?\n");
				return -1;
			}
		} else {
			dbg("Control's value not defined\n");
			return -1;
		}
		break;

	case SAHPI_CTRL_TYPE_DISCRETE:
		dbg("Discrete controls not supported\n");
		return -1;
	case SAHPI_CTRL_TYPE_ANALOG:
		dbg("Analog controls not supported\n");
		return -1;
	case SAHPI_CTRL_TYPE_STREAM:
		dbg("Stream controls not supported\n");
		return -1;
	case SAHPI_CTRL_TYPE_TEXT:
		dbg("Text controls not supported\n");
		return -1;
	case SAHPI_CTRL_TYPE_OEM:	
		dbg("Oem controls not supported\n");
		return -1;
        default:
		dbg("%s has invalid control state=%d\n", s->oid,working.Type);
                return -1;
        }

	memcpy(state,&working,sizeof(SaHpiCtrlStateT));
	return SA_OK;
}

static int snmp_bc_set_control_state(void *hnd, SaHpiResourceIdT id,
                                     SaHpiCtrlNumT num,
                                     SaHpiCtrlStateT *state)
{
        gchar *oid;
	int value;
        struct snmp_value set_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRdrT *rdr = oh_get_rdr_by_type(handle->rptcache, id, SAHPI_CTRL_RDR, num);
        struct ControlMibInfo *s =
                (struct ControlMibInfo *)oh_get_rdr_data(handle->rptcache, id, rdr->RecordId);

	if(state->Type != rdr->RdrTypeUnion.CtrlRec.Type) {
		dbg("Control %s type %d cannot be changed\n",s->oid,state->Type);
		return SA_ERR_HPI_INVALID_PARAMS;
	}

	switch (state->Type) {
	case SAHPI_CTRL_TYPE_DIGITAL:

		/* More icky dependencies on SaHpiStateDigitalT enum */
		switch (state->StateUnion.Digital) {
		case SAHPI_CTRL_STATE_OFF:
			value = s->digitalmap[SAHPI_CTRL_STATE_OFF];
			break;
		case SAHPI_CTRL_STATE_ON:
			value = s->digitalmap[SAHPI_CTRL_STATE_ON];
			break;	
		case SAHPI_CTRL_STATE_PULSE_OFF:
			value = s->digitalmap[SAHPI_CTRL_STATE_PULSE_OFF];
			break;
		case SAHPI_CTRL_STATE_PULSE_ON:
			value = s->digitalmap[SAHPI_CTRL_STATE_PULSE_ON];
			break;
		case SAHPI_CTRL_STATE_AUTO:
			value = s->digitalmap[ELEMENTS_IN_SaHpiStateDigitalT - 1];
			break;
		default:
			dbg("Spec Change: MAX_SaHpiStateDigitalT incorrect?\n");
			return -1;
		}

		if(value < 0) {
			dbg("Control state %d not allowed to be set\n",state->StateUnion.Digital);
			return SA_ERR_HPI_INVALID_CMD;
		}

		oid = snmp_derive_objid(rdr->Entity, s->oid);
		if(oid == NULL) {
			dbg("NULL SNMP OID returned for %s\n",s->oid);
			return -1;
		}

		set_value.type = ASN_INTEGER;
		set_value.integer = value;

		if((snmp_set(custom_handle->ss, oid, set_value) != 0)) {
			dbg("SNMP could not set %s; Type=%d.\n",s->oid,set_value.type);
			g_free(oid);
			return SA_ERR_HPI_NO_RESPONSE;
		}
		g_free(oid);

	case SAHPI_CTRL_TYPE_DISCRETE:
		dbg("Discrete controls not supported\n");
		return SA_ERR_HPI_INVALID_CMD;
	case SAHPI_CTRL_TYPE_ANALOG:
		dbg("Analog controls not supported\n");
		return SA_ERR_HPI_INVALID_CMD;
	case SAHPI_CTRL_TYPE_STREAM:
		dbg("Stream controls not supported\n");
		return SA_ERR_HPI_INVALID_CMD;
	case SAHPI_CTRL_TYPE_TEXT:
		dbg("Text controls not supported\n");
		return SA_ERR_HPI_INVALID_CMD;
	case SAHPI_CTRL_TYPE_OEM:	
		dbg("Oem controls not supported\n");
		return SA_ERR_HPI_INVALID_CMD;
        default:
		dbg("Request has invalid control state=%d\n", state->Type);
                return SA_ERR_HPI_INVALID_PARAMS;
        }
	
        return SA_OK;
}

#define get_inventory_data() \
do { \
        str_len = 0; \
        total_len = str_len; \
        if(s->oid.OidMfgDateTime != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidMfgDateTime); \
                if(oid == NULL) { \
                	working->MfgDateTime =  (SaHpiTimeT) SAHPI_TIME_UNSPECIFIED; \
                        dbg("NULL SNMP OID returned for MfgDateTime\n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                working->MfgDateTime = \
                                        g_strtod((char *)get_value.string, NULL); \
                        } else { \
                                dbg("%s Invalid type for Manufacturer inventory data\n",oid); \
                        } \
                } \
        } else { \
               	working->MfgDateTime =  (SaHpiTimeT) SAHPI_TIME_UNSPECIFIED; \
        } \
        total_len += str_len; \
        str_len = 0; \
        working->Manufacturer = (SaHpiTextBufferT *) base_ptr; \
        working->Manufacturer->DataLength = str_len; \
        working->Manufacturer->DataType = SAHPI_TL_TYPE_LANGUAGE; \
        working->Manufacturer->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidManufacturer != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidManufacturer); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for Manufacturer\n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                str_len = strlen((char *) get_value.string); \
                                working->Manufacturer->DataLength = str_len; \
                                strcpy((char *)working->Manufacturer->Data, (char *)get_value.string); \
                        } else { \
                                dbg("%s Invalid type for Manufacturer inventory data\n",oid); \
                        } \
                } \
        } \
        base_ptr = base_ptr + base_len + str_len; \
        working->ProductName = (SaHpiTextBufferT *)base_ptr; \
        total_len += str_len + base_len; \
        str_len = 0; \
	working->ProductName->DataLength = str_len; \
	working->ProductName->DataType = SAHPI_TL_TYPE_LANGUAGE; \
	working->ProductName->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidProductName != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidProductName); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for ProductName\n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                str_len = strlen((char *) get_value.string); \
                                working->ProductName->DataLength = str_len; \
                                strcpy((char *)working->ProductName->Data, (char *)get_value.string); \
                        } else { \
                                dbg("%s Invalid type for ProductName inventory data\n",oid); \
                        } \
                } \
        } \
        base_ptr = base_ptr + base_len + str_len; \
        working->ProductVersion = (SaHpiTextBufferT *)base_ptr; \
        total_len += str_len + base_len; \
        str_len = 0; \
	working->ProductVersion->DataLength = str_len; \
	working->ProductVersion->DataType = SAHPI_TL_TYPE_LANGUAGE; \
	working->ProductVersion->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidProductVersion != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidProductVersion); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for ProductVersion \n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_INTEGER )) { \
				int i = 0; \
				do { \
					conv_string[i++] = get_value.integer % 10 + '0'; \
				} while ((get_value.integer /= 10) > 0); \
				conv_string[i] = '\0'; \
                                str_len = sizeof(conv_string); \
                                working->ProductVersion->DataLength = str_len; \
                                strcpy((char *)working->ProductVersion->Data, (char *)(conv_string)); \
                        } else { \
                                dbg("%s Invalid type for ProductVersion inventory data\n",oid); \
                        } \
                } \
        } \
        base_ptr = base_ptr + base_len + str_len; \
        working->ModelNumber = (SaHpiTextBufferT *) base_ptr; \
        total_len += str_len + base_len; \
        str_len = 0; \
        working->ModelNumber->DataLength = str_len; \
	working->ModelNumber->DataType = SAHPI_TL_TYPE_LANGUAGE; \
	working->ModelNumber->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidModelNumber != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidModelNumber); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for ModelNumber \n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                str_len = strlen((char *) get_value.string); \
                                working->ModelNumber->DataLength = str_len; \
                                strcpy((char *)working->ModelNumber->Data, (char *)get_value.string); \
                        } else { \
                                dbg("%s Invalid type for ModelNumber inventory data\n",oid); \
                        } \
                } \
        } \
        base_ptr = base_ptr + base_len + str_len; \
        working->SerialNumber = (SaHpiTextBufferT *)base_ptr; \
        total_len += str_len + base_len; \
        str_len = 0; \
	working->SerialNumber->DataLength = str_len; \
	working->SerialNumber->DataType = SAHPI_TL_TYPE_LANGUAGE; \
	working->SerialNumber->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidSerialNumber != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidSerialNumber); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for OidSerialNumber \n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                str_len = strlen((char *) get_value.string); \
                                working->SerialNumber->DataLength = str_len; \
                                strcpy((char *)working->SerialNumber->Data, (char *)get_value.string); \
                        } else { \
                                dbg("%s Invalid type for SerialNumber inventory data\n",oid); \
                        } \
                } \
        } \
        base_ptr = base_ptr + base_len + str_len; \
        working->PartNumber = (SaHpiTextBufferT *)base_ptr; \
        total_len += str_len + base_len; \
        str_len = 0; \
	working->PartNumber->DataLength = str_len; \
	working->PartNumber->DataType = SAHPI_TL_TYPE_LANGUAGE; \
	working->PartNumber->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidPartNumber != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidPartNumber); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for OidPartNumber \n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                str_len = strlen((char *) get_value.string); \
                                working->PartNumber->DataLength = str_len; \
                                strcpy((char *)working->PartNumber->Data, (char *)get_value.string); \
                        } else { \
                                dbg("%s Invalid type for PartNumber inventory data\n",oid); \
                        } \
                } \
        } \
        base_ptr = base_ptr + base_len + str_len; \
        working->FileId = (SaHpiTextBufferT *)base_ptr; \
        total_len += str_len + base_len; \
        str_len = 0; \
	working->FileId->DataLength = str_len; \
	working->FileId->DataType = SAHPI_TL_TYPE_LANGUAGE; \
	working->FileId->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidFileId != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidFileId); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for OidFileId \n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                str_len = strlen((char *) get_value.string); \
                                working->FileId->DataLength = str_len; \
                                strcpy((char *)working->FileId->Data, (char *)get_value.string); \
                        } else { \
                                dbg("%s Invalid type for FileId inventory data\n",oid); \
                        } \
                } \
        } \
        base_ptr = base_ptr + base_len + str_len; \
        working->AssetTag = (SaHpiTextBufferT *)base_ptr; \
        total_len += str_len + base_len; \
        str_len = 0; \
        working->AssetTag->DataLength = str_len; \
        working->AssetTag->DataType = SAHPI_TL_TYPE_LANGUAGE; \
        working->AssetTag->Language = SAHPI_LANG_ENGLISH; \
        if(s->oid.OidAssetTag != NULL) { \
                oid = snmp_derive_objid(rdr->Entity,s->oid.OidAssetTag); \
                if(oid == NULL) { \
                        dbg("NULL SNMP OID returned for OidAssetTag \n"); \
                } else { \
                        rc = snmp_get(custom_handle->ss, oid, &get_value); \
                        if( (rc != 0) | \
                          !((get_value.type == ASN_INTEGER) | (get_value.type == ASN_OCTET_STR))) { \
                                dbg("SNMP could not read %s; Type=%d.\n",oid,get_value.type); \
                        } else if((rc == 0) && (get_value.type == ASN_OCTET_STR )) { \
                                str_len = strlen((char *) get_value.string); \
                                working->AssetTag->DataLength = str_len; \
                                strcpy((char *)working->AssetTag->Data, (char *)get_value.string); \
                        } else { \
                                dbg("%s Invalid type for AssetTag inventory data\n",oid); \
                        } \
                } \
        } \
        total_len += str_len + base_len; \
        str_len = 0; \
	working->CustomField[0] = NULL; \
} while(0)

#define init_inventory_buffer() \
do { \
        base_ptr = (uint *)l_data + sizeof(SaHpiInventoryDataT) + sizeof(SaHpiInventDataRecordT); \
} while(0) 
static int snmp_bc_get_inventory_info(void *hnd, SaHpiResourceIdT id,
                                      SaHpiEirIdT num,
                                      SaHpiInventoryDataT *data)
{
        int rc;
        SaHpiUint8T  str_len;
        size_t       base_len = sizeof(SaHpiTextTypeT) +sizeof(SaHpiLanguageT) +sizeof(SaHpiUint8T);
        size_t       total_len;
        uint         *base_ptr;
	char conv_string[20];
        SaHpiInventDataRecordTypeT l_recordtype;
        SaErrorT return_status = SA_OK;
        gchar *oid = NULL;
        SaHpiInventGeneralDataT *working;
        struct inventory_buffer *l_data;
        struct snmp_value get_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRdrT *rdr = oh_get_rdr_by_type(handle->rptcache, id, SAHPI_INVENTORY_RDR, num);

	total_len = 0;
        l_data = (struct inventory_buffer *) data;
        l_data->inventory_cntl.Validity =  SAHPI_INVENT_DATA_INVALID;
        l_data->inventory_cntl.DataRecords[0] = &l_data->inventory_fields;

        if(rdr != NULL) {

                struct InventoryMibInfo *s =
                        (struct InventoryMibInfo *)oh_get_rdr_data(handle->rptcache, id, rdr->RecordId);
                l_recordtype = s->inventory_type;

                 switch (l_recordtype) {
                         case SAHPI_INVENT_RECTYPE_INTERNAL_USE:
                                dbg("Not yet implemented SAHPI_INVENT_RECTYPE_INTERNAL_USE\n");
                                return_status = SA_ERR_HPI_UNKNOWN;
                                break;
                        case SAHPI_INVENT_RECTYPE_PRODUCT_INFO:
                        case SAHPI_INVENT_RECTYPE_BOARD_INFO:
                                working = (SaHpiInventGeneralDataT *)&l_data->inventory_cntl.DataRecords[0]->RecordData.ProductInfo;
                                init_inventory_buffer();
                                get_inventory_data();
                		l_data->inventory_cntl.Validity =  SAHPI_INVENT_DATA_VALID;
                                return_status = SA_OK;
                                break;
                        case SAHPI_INVENT_RECTYPE_CHASSIS_INFO:
                                working = (SaHpiInventGeneralDataT *)&l_data->inventory_cntl.DataRecords[0]->RecordData.ChassisInfo.GeneralData;
                                init_inventory_buffer();
                                l_data->inventory_cntl.DataRecords[0]->RecordData.ChassisInfo.Type = s->chassis_type;
                                get_inventory_data();
                		l_data->inventory_cntl.Validity =  SAHPI_INVENT_DATA_VALID;
                                return_status = SA_OK;
                                break;
                        case SAHPI_INVENT_RECTYPE_OEM:
                                dbg("Not yet implemented SAHPI_INVENT_RECTYPE_OEM\n");
                                return_status = SA_ERR_HPI_UNKNOWN;
                                break;
                        default:
                                dbg("Invalid inventory type %x in RDR\n", l_recordtype);
                                return_status = SA_ERR_HPI_UNKNOWN;
                                break;
                    }

                l_data->inventory_cntl.DataRecords[0]->RecordType = l_recordtype;
                l_data->inventory_cntl.DataRecords[0]->DataLength = total_len;

                /* FIXME:: Do we need to add events as well? */

        } else {
                dbg("No Inventory RDR found for ResourceId %x, EirId %x\n", id, num);
                return_status = SA_ERR_HPI_NOT_PRESENT;
        }

        return(return_status);

}

static int snmp_bc_get_inventory_size(void *hnd, SaHpiResourceIdT id,
                                      SaHpiEirIdT num,  /* yes, they don't call it a
                                                         * num, but it still is one
                                                        */
                                      size_t *size)
{
        SaErrorT rc, return_status = SA_OK;
        struct inventory_buffer *l_inventdata;
        l_inventdata =(struct inventory_buffer *)g_malloc(sizeof(struct inventory_buffer));
        if (l_inventdata != NULL) {
                rc = snmp_bc_get_inventory_info(hnd, id, num, &l_inventdata->inventory_cntl);
               *size = (l_inventdata->inventory_cntl.DataRecords[0]->DataLength +
                          sizeof(SaHpiInventoryDataT) +
                           sizeof(SaHpiInventDataRecordT));

                g_free(l_inventdata);
                return_status = SA_OK;
        } else {
                dbg("No space available ");
                return_status = SA_ERR_HPI_OUT_OF_SPACE ;
        }

        return return_status;
}


static int snmp_bc_set_inventory_info(void *hnd, SaHpiResourceIdT id,
                                      SaHpiEirIdT num,
                                      const SaHpiInventoryDataT *data)
{
        return -1;
}

static int snmp_bc_get_watchdog_info(void *hnd, SaHpiResourceIdT id,
                                     SaHpiWatchdogNumT num,
                                     SaHpiWatchdogT *wdt)
{
	/* Watchdog not supported */
        return SA_ERR_HPI_NOT_PRESENT;
}

static int snmp_bc_set_watchdog_info(void *hnd, SaHpiResourceIdT id,
                                     SaHpiWatchdogNumT num,
                                     SaHpiWatchdogT *wdt)
{
	/* Watchdog not supported */
        return SA_ERR_HPI_NOT_PRESENT;
}

static int snmp_bc_reset_watchdog(void *hnd, SaHpiResourceIdT id,
                                  SaHpiWatchdogNumT num)
{
 	/* Watchdog not supported */
        return SA_ERR_HPI_NOT_PRESENT;
}

static int snmp_bc_get_hotswap_state(void *hnd, SaHpiResourceIdT id,
				     SaHpiHsStateT *state)
{
	gchar *oid;
        struct snmp_value get_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRptEntryT *res = oh_get_resource_by_id(handle->rptcache, id);
        struct ResourceMibInfo *s =
                (struct ResourceMibInfo *)oh_get_resource_data(handle->rptcache, id);

	if(s->OidHealth == NULL) { return SA_ERR_HPI_INVALID_CMD; }

	oid = snmp_derive_objid(res->ResourceEntity, s->OidHealth);
	if(oid == NULL) {
		dbg("NULL SNMP OID returned for %s\n",s->OidHealth);
		return -1;
	}

	if((snmp_get(custom_handle->ss,oid,&get_value) == 0) && (get_value.type == ASN_INTEGER)) {
		if(get_value.integer == s->HealthyValue) { *state = SAHPI_HS_STATE_ACTIVE_HEALTHY; }
		else { *state = SAHPI_HS_STATE_ACTIVE_UNHEALTHY; }
        } else {
		dbg("Couldn't fetch SNMP %s vector; Type=%d\n",oid,get_value.type);
		g_free(oid);
		return -1;
	}

	g_free(oid);
        return SA_OK;
}

static int snmp_bc_set_hotswap_state(void *hnd, SaHpiResourceIdT id,
				     SaHpiHsStateT state)
{
        return SA_ERR_HPI_UNSUPPORTED_API;
}

static int snmp_bc_request_hotswap_action(void *hnd, SaHpiResourceIdT id,
				          SaHpiHsActionT act)
{
        return SA_ERR_HPI_UNSUPPORTED_API;
}

static int snmp_bc_get_reset_state(void *hnd, SaHpiResourceIdT id,
			           SaHpiResetActionT *act)
{
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct ResourceMibInfo *s =
                (struct ResourceMibInfo *)oh_get_resource_data(handle->rptcache, id);

	if(s->OidReset == NULL) { return SA_ERR_HPI_INVALID_CMD; }

	*act = SAHPI_RESET_DEASSERT;

	return SA_OK;
}

static int snmp_bc_set_reset_state(void *hnd, SaHpiResourceIdT id,
			           SaHpiResetActionT act)
{
	gchar *oid;
        struct snmp_value set_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRptEntryT *res = oh_get_resource_by_id(handle->rptcache, id);
        struct ResourceMibInfo *s =
                (struct ResourceMibInfo *)oh_get_resource_data(handle->rptcache, id);

	if(s->OidReset == NULL) { return SA_ERR_HPI_INVALID_CMD; }

	switch (act) {
	case SAHPI_RESET_ASSERT: /* RESET_ASSERT = RESET_DEASSERT Action */
	case SAHPI_RESET_DEASSERT:
		return SA_ERR_HPI_INVALID_CMD;
	case SAHPI_COLD_RESET: /* COLD = WARM Reset Action */
	case SAHPI_WARM_RESET:
		oid = snmp_derive_objid(res->ResourceEntity, s->OidReset);
		if(oid == NULL) {
			dbg("NULL SNMP OID returned for %s\n",s->OidReset);
			return -1;
		}
		
		set_value.type = ASN_INTEGER;
		set_value.integer = 1;
		
		if((snmp_set(custom_handle->ss, oid, set_value) != 0)) {
			dbg("SNMP could not set %s; Type=%d.\n",s->OidReset,set_value.type);
			g_free(oid);
			return SA_ERR_HPI_NO_RESPONSE;
		}
		g_free(oid);
		break;
	default:
		dbg("Invalid Reset Action Type - %d\n", act);
		return SA_ERR_HPI_INVALID_PARAMS;
	}

        return SA_OK;
}

static int snmp_bc_get_power_state(void *hnd, SaHpiResourceIdT id,
			           SaHpiHsPowerStateT *state)
{
	gchar *oid;
	int rtn_code = SA_OK;
        struct snmp_value get_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRptEntryT *res = oh_get_resource_by_id(handle->rptcache, id);
        struct ResourceMibInfo *s =
                (struct ResourceMibInfo *)oh_get_resource_data(handle->rptcache, id);

	if(s->OidPowerState == NULL) { return SA_ERR_HPI_INVALID_CMD; }

	oid = snmp_derive_objid(res->ResourceEntity, s->OidPowerState);
	if(oid == NULL) {
		dbg("NULL SNMP OID returned for %s\n",s->OidPowerState);
		return -1;
	}

	if((snmp_get(custom_handle->ss,oid,&get_value) == 0) && (get_value.type == ASN_INTEGER)) {
		switch (get_value.integer) {
		case 0:
			*state = SAHPI_HS_POWER_OFF;
			break;
		case 1:
			*state = SAHPI_HS_POWER_ON;
			break;
		default:
			dbg("Invalid power state read for oid=%s\n",oid);
			rtn_code = -1;
		}
        } else {
		dbg("Couldn't fetch SNMP %s vector; Type=%d\n",oid,get_value.type);
		rtn_code = -1;
	}

	g_free(oid);
        return rtn_code;
}

static int snmp_bc_set_power_state(void *hnd, SaHpiResourceIdT id,
			           SaHpiHsPowerStateT state)
{
	gchar *oid;
	int rtn_code = SA_OK;
        struct snmp_value set_value;
        struct oh_handler_state *handle = (struct oh_handler_state *)hnd;
        struct snmp_bc_hnd *custom_handle = (struct snmp_bc_hnd *)handle->data;
        SaHpiRptEntryT *res = oh_get_resource_by_id(handle->rptcache, id);
        struct ResourceMibInfo *s =
                (struct ResourceMibInfo *)oh_get_resource_data(handle->rptcache, id);

	if(s->OidPowerOnOff == NULL) { return SA_ERR_HPI_INVALID_CMD; }

	oid = snmp_derive_objid(res->ResourceEntity, s->OidPowerOnOff);
	if(oid == NULL) {
		dbg("NULL SNMP OID returned for %s\n",s->OidPowerOnOff);
		return -1;
	}

	set_value.type = ASN_INTEGER;
	switch (state) {
	case SAHPI_HS_POWER_OFF:
		set_value.integer = 0;
		if((snmp_set(custom_handle->ss, oid, set_value) != 0)) {
			dbg("SNMP could not set %s; Type=%d.\n",s->OidPowerOnOff,set_value.type);
			rtn_code = SA_ERR_HPI_NO_RESPONSE;
		}
		break;
		
	case SAHPI_HS_POWER_ON:
		set_value.integer = 1;
		if((snmp_set(custom_handle->ss, oid, set_value) != 0)) {
			dbg("SNMP could not set %s; Type=%d.\n",s->OidPowerOnOff,set_value.type);
			rtn_code = SA_ERR_HPI_NO_RESPONSE;
		}
		break;
		
	case SAHPI_HS_POWER_CYCLE:
	        {
			SaHpiResetActionT act = SAHPI_COLD_RESET;
			rtn_code=snmp_bc_set_reset_state(hnd, id, act);
	        }
		break;
	default:
		dbg("Invalid Power Action Type - %d\n", state);
		rtn_code = SA_ERR_HPI_INVALID_PARAMS;
	}

	g_free(oid);
        return rtn_code;
}

static int snmp_bc_get_indicator_state(void *hnd, SaHpiResourceIdT id,
				       SaHpiHsIndicatorStateT *state)
{
        return SA_ERR_HPI_INVALID_CMD;
}

static int snmp_bc_set_indicator_state(void *hnd, SaHpiResourceIdT id,
				       SaHpiHsIndicatorStateT state)
{
        return SA_ERR_HPI_INVALID_CMD;
}

static int snmp_bc_control_parm(void *hnd, SaHpiResourceIdT id, SaHpiParmActionT act)
{
        return -1;
}

struct oh_abi_v2 oh_snmp_bc_plugin = {
        .open				= snmp_bc_open,
        .close				= snmp_bc_close,
        .get_event			= snmp_bc_get_event,
        .discover_resources     	= snmp_bc_discover_resources,
        .set_resource_tag               = snmp_bc_set_resource_tag,
        .set_resource_severity          = snmp_bc_set_resource_severity,
        .get_self_id			= snmp_bc_get_self_id,
        .get_sel_info			= snmp_bc_get_sel_info,
        .set_sel_time			= snmp_bc_set_sel_time,
        .add_sel_entry			= snmp_bc_add_sel_entry,
        .del_sel_entry			= snmp_bc_del_sel_entry,
        .get_sel_entry			= snmp_bc_get_sel_entry,
        .get_sensor_data		= snmp_bc_get_sensor_data,
        .get_sensor_thresholds		= snmp_bc_get_sensor_thresholds,
        .set_sensor_thresholds		= snmp_bc_set_sensor_thresholds,
        .get_sensor_event_enables	= snmp_bc_get_sensor_event_enables,
        .set_sensor_event_enables	= snmp_bc_set_sensor_event_enables,
        .get_control_state		= snmp_bc_get_control_state,
        .set_control_state		= snmp_bc_set_control_state,
        .get_inventory_size		= snmp_bc_get_inventory_size,
        .get_inventory_info		= snmp_bc_get_inventory_info,
        .set_inventory_info		= snmp_bc_set_inventory_info,
        .get_watchdog_info		= snmp_bc_get_watchdog_info,
        .set_watchdog_info		= snmp_bc_set_watchdog_info,
        .reset_watchdog			= snmp_bc_reset_watchdog,
        .get_hotswap_state		= snmp_bc_get_hotswap_state,
        .set_hotswap_state		= snmp_bc_set_hotswap_state,
        .request_hotswap_action		= snmp_bc_request_hotswap_action,
        .get_power_state		= snmp_bc_get_power_state,
        .set_power_state		= snmp_bc_set_power_state,
        .get_indicator_state		= snmp_bc_get_indicator_state,
        .set_indicator_state		= snmp_bc_set_indicator_state,
        .control_parm			= snmp_bc_control_parm,
        .get_reset_state		= snmp_bc_get_reset_state,
        .set_reset_state		= snmp_bc_set_reset_state
};

int get_interface(void **pp, const uuid_t uuid)
{
        if(uuid_compare(uuid, UUID_OH_ABI_V2)==0) {
                *pp = &oh_snmp_bc_plugin;
                return 0;
        }

        *pp = NULL;
        return -1;
}
